import copy
from typing import Optional, Tuple, Union

import torch
import torch.nn as nn
import torch.nn.functional as F
from funasr.models.transformer.utils.nets_utils import make_pad_mask


def sense_voice_encode_forward(
	self,
	x: torch.Tensor,
	ilens: torch.Tensor = None,
	**kwargs,
):
	use_padmask = self.use_padmask
	x = F.gelu(self.conv1(x))
	x = F.gelu(self.conv2(x))
	x = x.permute(0, 2, 1)
	
	n_frames = x.size(1)
	max_pos = self.positional_embedding.size(0)
	max_pos = n_frames if n_frames < max_pos else max_pos
	x = (x[:, :max_pos, :] + self.positional_embedding[None, :max_pos, :]).to(x.dtype)
	
	
	if ilens is not None:
		if self.downsample_rate == 4:
			olens = (
				1
				+ (
					ilens
					- self.conv1.kernel_size[0]
					+ 2 * self.conv1.padding[0]
				)
				// self.conv1.stride[0]
			)
		else:
			olens = ilens
		olens = (
			1
			+ (
				olens
				- self.conv2.kernel_size[0]
				+ 2 * self.conv2.padding[0]
			)
			// self.conv2.stride[0]
		)
		olens = torch.clamp(olens, max=max_pos)
	else:
		olens = None
	
	if use_padmask and olens is not None:
		padding_mask = (~make_pad_mask(olens)[:, None, :]).to(x.device)
	else:
		padding_mask = None
	
	for layer, block in enumerate(self.blocks):
		x = block(x, mask=padding_mask, is_pad_mask=True)
		

	x = self.ln_post(x)
	
	if ilens is None:
		return x
	else:
		return x, olens



from dataclasses import dataclass
from typing import Dict
from typing import Iterable, Optional
import types
import time
import numpy as np
import torch
import torch.nn.functional as F
from torch import Tensor
from torch import nn
from torch.cuda.amp import autocast
from funasr.metrics.compute_acc import compute_accuracy
from funasr.losses.label_smoothing_loss import LabelSmoothingLoss
from funasr.train_utils.device_funcs import force_gatherable
from . import whisper_lib as whisper
from funasr.utils.load_utils import load_audio_text_image_video, extract_fbank

from funasr.register import tables




@tables.register("model_classes", "SenseVoice")
class SenseVoice(nn.Module):
    def __init__(self, *args, **kwargs):
        super().__init__()
        
        dims = kwargs.get("dims", {})
        dims = whisper.model.ModelDimensions(**dims)
        model = whisper.model.Whisper(dims=dims)
        
        # encoder
        model.encoder.downsample_rate = kwargs.get("downsample_rate", 4)
        model.encoder.use_padmask = kwargs.get("use_padmask", True)
        from .encoder import sense_voice_encode_forward
        model.encoder.forward = types.MethodType(sense_voice_encode_forward, model.encoder)
        
        # decoder
        model.decoder.use_padmask = kwargs.get("use_padmask", True)
        from .decoder import sense_voice_decode_forward
        model.decoder.forward = types.MethodType(sense_voice_decode_forward, model.decoder)
        
        self.model = model
        
        self.encoder_output_size = self.model.dims.n_audio_state
        
        self.activation_checkpoint = kwargs.get("activation_checkpoint", False)
        self.ignore_id = kwargs.get("ignore_id", -1)
        self.vocab_size = kwargs.get("vocab_size", -1)
        self.length_normalized_loss = kwargs.get("length_normalized_loss", True)
        self.criterion_att = LabelSmoothingLoss(
            size=self.vocab_size,
            padding_idx=self.ignore_id,
            smoothing=kwargs.get("lsm_weight", 0.0),
            normalize_length=self.length_normalized_loss,
        )
        
        specaug = kwargs.get("specaug", None)
        if specaug is not None:
            specaug_class = tables.specaug_classes.get(specaug)
            specaug = specaug_class(**kwargs.get("specaug_conf", {}))
        self.specaug = specaug

 
    def forward(
        self,
        speech: torch.Tensor,
        speech_lengths: torch.Tensor,
        text: torch.Tensor,
        text_lengths: torch.Tensor,
        **kwargs,
    ):
        target_mask = kwargs.get("target_mask", None)
    
        # import pdb;
        # pdb.set_trace()
        if len(text_lengths.size()) > 1:
            text_lengths = text_lengths[:, 0]
        if len(speech_lengths.size()) > 1:
            speech_lengths = speech_lengths[:, 0]
    
        batch_size = speech.shape[0]

        if self.activation_checkpoint:
            from torch.utils.checkpoint import checkpoint
            encoder_out, encoder_out_lens = checkpoint(self.encode, speech, speech_lengths, use_reentrant=False)
        else:
            encoder_out, encoder_out_lens = self.encode(speech, speech_lengths)

        loss_att, acc_att, cer_att, wer_att = self._calc_att_loss(
            encoder_out, encoder_out_lens, text, text_lengths, target_mask=target_mask
        )
        loss = loss_att
        stats = {}
        stats["acc"] = acc_att
        stats["loss"] = torch.clone(loss.detach())
        stats["batch_size"] = batch_size
        
        # force_gatherable: to-device and to-tensor if scalar for DataParallel
        if self.length_normalized_loss:
            batch_size = int((text_lengths + 1).sum())
        loss, stats, weight = force_gatherable((loss, stats, batch_size), loss.device)
        return loss, stats, weight

    def encode(
        self, speech: torch.Tensor, speech_lengths: torch.Tensor, **kwargs,
    ) :
        """Encoder. Note that this method is used by asr_inference.py
        Args:
                speech: (Batch, Length, ...)
                speech_lengths: (Batch, )
                ind: int
        """
        with autocast(False):

            # Data augmentation
            if self.specaug is not None and self.training:
                speech, speech_lengths = self.specaug(speech, speech_lengths)


        # Forward encoder
        encoder_out, encoder_out_lens = self.model.encoder(speech.permute(0, 2, 1), speech_lengths)
    
        return encoder_out, encoder_out_lens


    def _calc_att_loss(
            self,
            encoder_out: torch.Tensor,
            encoder_out_lens: torch.Tensor,
            ys_pad: torch.Tensor,
            ys_pad_lens: torch.Tensor,
            **kwargs,
    ):
        target_mask = kwargs.get("target_mask", None)
        stats = {}
        
        # 1. Forward decoder
        decoder_out = self.model.decoder(
            x=ys_pad, xa=encoder_out, hlens=encoder_out_lens, ys_in_lens=ys_pad_lens
        )
        
        # 2. Compute attention loss
        mask = torch.ones_like(ys_pad) * (-1)
        ys_pad_mask = (ys_pad * target_mask + mask * (1-target_mask)).to(torch.int64)
        ys_pad_mask[ys_pad_mask == 0] = -1
        loss_att = self.criterion_att(decoder_out[:, :-1, :], ys_pad_mask[:, 1:])

        with torch.no_grad():
            preds = torch.argmax(decoder_out, -1)
            acc_att = compute_accuracy(preds[:, :-1], ys_pad_mask[:, 1:], ignore_label=self.ignore_id)

        return loss_att, acc_att, None, None


    def inference(self,
                  data_in,
                  data_lengths=None,
                  key: list = None,
                  tokenizer=None,
                  frontend=None,
                  **kwargs,
                  ):
        if kwargs.get("batch_size", 1) > 1:
            raise NotImplementedError("batch decoding is not implemented")

        if frontend is None and not hasattr(self, "frontend"):
            frontend_class = tables.frontend_classes.get("WhisperFrontend")
            frontend = frontend_class(n_mels=self.model.dims.n_mels, do_pad_trim=kwargs.get("do_pad_trim", True))
            self.frontend = frontend
        else:
            frontend = frontend if frontend is not None else self.frontend

        meta_data = {}
        if isinstance(data_in, torch.Tensor) and kwargs.get("data_type", "sound") == "fbank":  # fbank
            speech, speech_lengths = data_in, data_lengths
            if len(speech.shape) < 3:
                speech = speech[None, :, :]
            if speech_lengths is None:
                speech_lengths = speech.shape[1]
        else:
            # extract fbank feats
            time1 = time.perf_counter()
            audio_sample_list = load_audio_text_image_video(data_in, fs=frontend.fs if hasattr(frontend, "fs") else 16000, audio_fs=kwargs.get("fs", 16000),
                                                            data_type=kwargs.get("data_type", "sound"),
                                                            tokenizer=tokenizer)
            time2 = time.perf_counter()
            meta_data["load_data"] = f"{time2 - time1:0.3f}"
            speech, speech_lengths = extract_fbank(audio_sample_list, data_type=kwargs.get("data_type", "sound"),
                                                   frontend=frontend)
            time3 = time.perf_counter()
            meta_data["extract_feat"] = f"{time3 - time2:0.3f}"
            frame_shift = frontend.frame_shift if hasattr(frontend, "frame_shift") else 10
            lfr_n = frontend.lfr_n if hasattr(frontend, "lfr_n") else 1
            meta_data["batch_data_time"] = speech_lengths.sum().item() * frame_shift * lfr_n / 1000

        speech = speech.to(device=kwargs["device"])[0, :, :]
        speech_lengths = speech_lengths.to(device=kwargs["device"])

        DecodingOptions = kwargs.get("DecodingOptions", {})
        task = DecodingOptions.get("task", "ASR")
        if isinstance(task, str):
            task = [task]
        task = "".join([f"<|{x}|>" for x in task])
        initial_prompt = kwargs.get("initial_prompt", f"<|startoftranscript|>{task}")
        DecodingOptions["initial_prompt"] = initial_prompt
        
        language = DecodingOptions.get("language", None)
        language = None if language == "auto" else language
        DecodingOptions["language"] = language

        DecodingOptions["vocab_path"] = kwargs["tokenizer_conf"].get("vocab_path", None)
        
        
        if "without_timestamps" not in DecodingOptions:
            DecodingOptions["without_timestamps"] = True

    
        options = whisper.DecodingOptions(**DecodingOptions)
        
        result = whisper.decode(self.model, speech, options)
        text = f"{result.text}"
        results = []
        result_i = {"key": key[0], "text": text}

        results.append(result_i)
    
        return results, meta_data


@tables.register("model_classes", "SenseVoiceRWKV")
class SenseVoiceRWKV(nn.Module):
    def __init__(self, *args, **kwargs):
        super().__init__()
        
        dims = kwargs.get("dims", {})
        dims = whisper.model.ModelDimensions(**dims)
        model = whisper.model.Whisper(dims=dims)
        
        # encoder
        model.encoder.downsample_rate = kwargs.get("downsample_rate", 4)
        model.encoder.use_padmask = kwargs.get("use_padmask", True)
        from .encoder import sense_voice_encode_forward
        model.encoder.forward = types.MethodType(sense_voice_encode_forward, model.encoder)
        
        # decoder
        del model.decoder
        decoder = kwargs.get("decoder", "SenseVoiceDecoder")
        decoder_class = tables.decoder_classes.get(decoder)
        decoder = decoder_class(n_vocab=dims.n_vocab,
                                n_ctx=dims.n_text_ctx,
                                n_state=dims.n_text_state,
                                n_head=dims.n_text_head,
                                n_layer=dims.n_text_layer,
                                **kwargs.get("decoder_conf"))
        model.decoder = decoder
        
        self.model = model
        
        self.encoder_output_size = self.model.dims.n_audio_state
        
        self.activation_checkpoint = kwargs.get("activation_checkpoint", False)
        self.ignore_id = kwargs.get("ignore_id", -1)
        self.vocab_size = kwargs.get("vocab_size", -1)
        self.length_normalized_loss = kwargs.get("length_normalized_loss", True)
        self.criterion_att = LabelSmoothingLoss(
            size=self.vocab_size,
            padding_idx=self.ignore_id,
            smoothing=kwargs.get("lsm_weight", 0.0),
            normalize_length=self.length_normalized_loss,
        )
        
        specaug = kwargs.get("specaug", None)
        if specaug is not None:
            specaug_class = tables.specaug_classes.get(specaug)
            specaug = specaug_class(**kwargs.get("specaug_conf", {}))
        self.specaug = specaug
    
    def forward(
        self,
        speech: torch.Tensor,
        speech_lengths: torch.Tensor,
        text: torch.Tensor,
        text_lengths: torch.Tensor,
        **kwargs,
    ):
        target_mask = kwargs.get("target_mask", None)
        
        # import pdb;
        # pdb.set_trace()
        if len(text_lengths.size()) > 1:
            text_lengths = text_lengths[:, 0]
        if len(speech_lengths.size()) > 1:
            speech_lengths = speech_lengths[:, 0]
        
        batch_size = speech.shape[0]
        
        if self.activation_checkpoint:
            from torch.utils.checkpoint import checkpoint
            encoder_out, encoder_out_lens = checkpoint(self.encode, speech, speech_lengths, use_reentrant=False)
        else:
            encoder_out, encoder_out_lens = self.encode(speech, speech_lengths)
        
        loss_att, acc_att, cer_att, wer_att = self._calc_att_loss(
            encoder_out, encoder_out_lens, text, text_lengths, target_mask=target_mask
        )
        loss = loss_att
        stats = {}
        stats["acc"] = acc_att
        stats["loss"] = torch.clone(loss.detach())
        stats["batch_size"] = batch_size
        
        # force_gatherable: to-device and to-tensor if scalar for DataParallel
        if self.length_normalized_loss:
            batch_size = int((text_lengths + 1).sum())
        loss, stats, weight = force_gatherable((loss, stats, batch_size), loss.device)
        return loss, stats, weight
    
    def encode(
        self, speech: torch.Tensor, speech_lengths: torch.Tensor, **kwargs,
    ):
        """Encoder. Note that this method is used by asr_inference.py
        Args:
                speech: (Batch, Length, ...)
                speech_lengths: (Batch, )
                ind: int
        """
        with autocast(False):
            # Data augmentation
            if self.specaug is not None and self.training:
                speech, speech_lengths = self.specaug(speech, speech_lengths)
        
        # Forward encoder
        encoder_out, encoder_out_lens = self.model.encoder(speech.permute(0, 2, 1), speech_lengths)
        
        return encoder_out, encoder_out_lens
    
    def _calc_att_loss(
        self,
        encoder_out: torch.Tensor,
        encoder_out_lens: torch.Tensor,
        ys_pad: torch.Tensor,
        ys_pad_lens: torch.Tensor,
        **kwargs,
    ):
        target_mask = kwargs.get("target_mask", None)
        stats = {}
        
        # 1. Forward decoder
        decoder_out = self.model.decoder(
            x=ys_pad, xa=encoder_out, hlens=encoder_out_lens, ys_in_lens=ys_pad_lens
        )
        
        # 2. Compute attention loss
        mask = torch.ones_like(ys_pad) * (-1)
        ys_pad_mask = (ys_pad * target_mask + mask * (1 - target_mask)).to(torch.int64)
        ys_pad_mask[ys_pad_mask == 0] = -1
        loss_att = self.criterion_att(decoder_out[:, :-1, :], ys_pad_mask[:, 1:])
        
        with torch.no_grad():
            preds = torch.argmax(decoder_out, -1)
            acc_att = compute_accuracy(preds[:, :-1], ys_pad_mask[:, 1:], ignore_label=self.ignore_id)
        
        return loss_att, acc_att, None, None
    
    def inference(self,
                  data_in,
                  data_lengths=None,
                  key: list = None,
                  tokenizer=None,
                  frontend=None,
                  **kwargs,
                  ):
        if kwargs.get("batch_size", 1) > 1:
            raise NotImplementedError("batch decoding is not implemented")
        
        if frontend is None and not hasattr(self, "frontend"):
            frontend_class = tables.frontend_classes.get("WhisperFrontend")
            frontend = frontend_class(n_mels=self.model.dims.n_mels, do_pad_trim=kwargs.get("do_pad_trim", True))
            self.frontend = frontend
        else:
            frontend = frontend if frontend is not None else self.frontend
        
        meta_data = {}
        if isinstance(data_in, torch.Tensor) and kwargs.get("data_type", "sound") == "fbank":  # fbank
            speech, speech_lengths = data_in, data_lengths
            if len(speech.shape) < 3:
                speech = speech[None, :, :]
            if speech_lengths is None:
                speech_lengths = speech.shape[1]
        else:
            # extract fbank feats
            time1 = time.perf_counter()
            audio_sample_list = load_audio_text_image_video(data_in,
                                                            fs=frontend.fs if hasattr(frontend, "fs") else 16000,
                                                            audio_fs=kwargs.get("fs", 16000),
                                                            data_type=kwargs.get("data_type", "sound"),
                                                            tokenizer=tokenizer)
            time2 = time.perf_counter()
            meta_data["load_data"] = f"{time2 - time1:0.3f}"
            speech, speech_lengths = extract_fbank(audio_sample_list, data_type=kwargs.get("data_type", "sound"),
                                                   frontend=frontend)
            time3 = time.perf_counter()
            meta_data["extract_feat"] = f"{time3 - time2:0.3f}"
            frame_shift = frontend.frame_shift if hasattr(frontend, "frame_shift") else 10
            lfr_n = frontend.lfr_n if hasattr(frontend, "lfr_n") else 1
            meta_data["batch_data_time"] = speech_lengths.sum().item() * frame_shift * lfr_n / 1000
        
        speech = speech.to(device=kwargs["device"])[0, :, :]
        speech_lengths = speech_lengths.to(device=kwargs["device"])
        
        DecodingOptions = kwargs.get("DecodingOptions", {})
        task = DecodingOptions.get("task", "ASR")
        if isinstance(task, str):
            task = [task]
        task = "".join([f"<|{x}|>" for x in task])
        initial_prompt = kwargs.get("initial_prompt", f"<|startoftranscript|>{task}")
        DecodingOptions["initial_prompt"] = initial_prompt
        
        language = DecodingOptions.get("language", None)
        language = None if language == "auto" else language
        DecodingOptions["language"] = language
        
        DecodingOptions["vocab_path"] = kwargs["tokenizer_conf"].get("vocab_path", None)
        
        if "without_timestamps" not in DecodingOptions:
            DecodingOptions["without_timestamps"] = True
        
        options = whisper.DecodingOptions(**DecodingOptions)
        
        result = whisper.decode(self.model, speech, options)
        text = f"{result.text}"
        results = []
        result_i = {"key": key[0], "text": text}
        
        results.append(result_i)
        
        return results, meta_data



import copy
from typing import Optional, Tuple, Union

import torch
import torch.nn as nn
import torch.nn.functional as F
from funasr.models.transformer.utils.nets_utils import make_pad_mask
from funasr.register import tables
import base64
import gzip
from dataclasses import dataclass
from typing import Dict, Iterable, Optional

import numpy as np
import torch
import torch.nn.functional as F
from torch import Tensor, nn


class LayerNorm(nn.LayerNorm):
    def forward(self, x: Tensor) -> Tensor:
        return super().forward(x.float()).type(x.dtype)


class Linear(nn.Linear):
    def forward(self, x: Tensor) -> Tensor:
        return F.linear(
            x,
            self.weight.to(x.dtype),
            None if self.bias is None else self.bias.to(x.dtype),
        )



def sense_voice_decode_forward(
	self,
	x: torch.Tensor,
	xa: torch.Tensor,
	kv_cache: Optional[dict] = None,
	**kwargs,
):
	"""Forward decoder.

	Args:
		hs_pad: encoded memory, float32  (batch, maxlen_in, feat)
		hlens: (batch)
		ys_in_pad:
			input token ids, int64 (batch, maxlen_out)
			if input_layer == "embed"
			input tensor (batch, maxlen_out, #mels) in the other cases
		ys_in_lens: (batch)
	Returns:
		(tuple): tuple containing:

		x: decoded token score before softmax (batch, maxlen_out, token)
			if use_output_layer is True,
		olens: (batch, )
	"""
	# import pdb;pdb.set_trace()
	use_padmask = self.use_padmask
	hlens = kwargs.get("hlens", None)
	
	ys_in_lens = kwargs.get("ys_in_lens", None)
	
	offset = next(iter(kv_cache.values())).shape[1] if kv_cache else 0
	tgt, memory = x, xa
	tgt[tgt == -1] = 0
	tgt = (
		self.token_embedding(tgt)
		+ self.positional_embedding[offset: offset + tgt.size(1)]
	)
	# tgt = self.dropout(tgt)
	
	x = tgt.to(memory.dtype)
	
	if use_padmask and hlens is not None:
		memory_mask = (~make_pad_mask(hlens)[:, None, :]).to(memory.device)
	else:
		memory_mask = None
	
	for layer, block in enumerate(self.blocks):
		x = block(x, memory, mask=self.mask, memory_mask=memory_mask, is_pad_mask=False, is_pad_memory_mask=True)
	
	x = self.ln(x)
	x = (
		x @ torch.transpose(self.token_embedding.weight.to(x.dtype), 0, 1)
	).float()
	
	return x


class MultiHeadAttention(nn.Module):
	def __init__(self, n_state: int, n_head: int):
		super().__init__()
		self.n_head = n_head
		self.query = Linear(n_state, n_state)
		self.key = Linear(n_state, n_state, bias=False)
		self.value = Linear(n_state, n_state)
		self.out = Linear(n_state, n_state)
	
	def forward(
		self,
		x: Tensor,
		xa: Optional[Tensor] = None,
		mask: Optional[Tensor] = None,
		kv_cache: Optional[dict] = None,
		**kwargs,
	):
		is_pad_mask = kwargs.get("is_pad_mask", False)
		
		q = self.query(x)
		
		if kv_cache is None or xa is None or self.key not in kv_cache:
			# hooks, if installed (i.e. kv_cache is not None), will prepend the cached kv tensors;
			# otherwise, perform key/value projections for self- or cross-attention as usual.
			k = self.key(x if xa is None else xa)
			v = self.value(x if xa is None else xa)
		else:
			# for cross-attention, calculate keys and values once and reuse in subsequent calls.
			k = kv_cache[self.key]
			v = kv_cache[self.value]
		
		wv, qk = self.qkv_attention(q, k, v, mask, is_pad_mask=is_pad_mask)
		return self.out(wv), qk
	
	def qkv_attention(
		self, q: Tensor, k: Tensor, v: Tensor, mask: Optional[Tensor] = None, **kwargs,
	):
		is_pad_mask = kwargs.get("is_pad_mask", False)
		n_batch, n_ctx, n_state = q.shape
		scale = (n_state // self.n_head) ** -0.25
		q = q.view(*q.shape[:2], self.n_head, -1).permute(0, 2, 1, 3) * scale
		k = k.view(*k.shape[:2], self.n_head, -1).permute(0, 2, 3, 1) * scale
		v = v.view(*v.shape[:2], self.n_head, -1).permute(0, 2, 1, 3)
		
		qk = q @ k
		if mask is not None:
			if not is_pad_mask:
				qk = qk + mask[:n_ctx, :n_ctx]
			else:
				mask = mask.unsqueeze(1).eq(0)  # (batch, 1, *, time2)
				min_value = float(
					np.finfo(torch.tensor(0, dtype=qk.dtype).numpy().dtype).min
				)
				qk = qk.masked_fill(mask, min_value)
		
		qk = qk.float()
		
		w = F.softmax(qk, dim=-1).to(q.dtype)
		if mask is not None and is_pad_mask:
			w = w.masked_fill(mask, 0.0)
		return (w @ v).permute(0, 2, 1, 3).flatten(start_dim=2), qk.detach()


from funasr.models.sense_voice.rwkv_v6 import RWKVLayer
from omegaconf import OmegaConf


class ResidualAttentionBlockRWKV(nn.Module):
	def __init__(self, n_state: int, n_head: int, cross_attention: bool = False, layer_id=0, **kwargs):
		super().__init__()
		
		rwkv_cfg = kwargs.get("rwkv_cfg", {})
		args = OmegaConf.create(rwkv_cfg)
		self.attn = RWKVLayer(args=args, layer_id=layer_id)
		if args.get("datatype", "bf16") == "bf16":
			self.attn.to(torch.bfloat16)
			
		self.ln0 = None
		if layer_id == 0 and not args.get("ln0", True):
			self.ln0 = LayerNorm(args.n_embd)
			if args.get("init_rwkv", True):
				print("init_rwkv")
				layer_id = 0
				scale = ((1 + layer_id) / args.get("n_layer")) ** 0.7
				nn.init.constant_(self.ln0.weight, scale)
		self.layer_id = layer_id
		self.args = args
		
		self.ln1 = None
		if not args.get("ln1", True):
			self.ln1 = LayerNorm(args.n_embd)
			# init
			if args.get("init_rwkv", True):
				print("init_rwkv")
				scale = ((1 + layer_id) / args.get("n_layer")) ** 0.7
				nn.init.constant_(self.ln1.weight, scale)
		
		self.cross_attn = (
			MultiHeadAttention(n_state, n_head) if cross_attention else None
		)
		self.cross_attn_ln = LayerNorm(n_state) if cross_attention else None
		
		n_mlp = n_state * 4
		self.mlp = nn.Sequential(
			Linear(n_state, n_mlp), nn.GELU(), Linear(n_mlp, n_state)
		)
		self.mlp_ln = LayerNorm(n_state)
	
	def forward(
		self,
		x: Tensor,
		xa: Optional[Tensor] = None,
		mask: Optional[Tensor] = None,
		kv_cache: Optional[dict] = None,
		**kwargs,
	):
		is_pad_mask = kwargs.get("is_pad_mask", False)
		is_pad_memory_mask = kwargs.get("is_pad_memory_mask", False)
		
		if self.layer_id == 0 and self.ln0 is not None:
			x = self.ln0(x)
		
		if self.ln1 is None:
			x = x + self.attn(x, mask=mask, kv_cache=kv_cache, is_pad_mask=is_pad_mask)[0]
		else:
			x = x + self.attn(self.ln1(x), mask=mask, kv_cache=kv_cache, is_pad_mask=is_pad_mask)[0]
	
		if self.cross_attn:
			x = x + self.cross_attn(self.cross_attn_ln(x), xa, kv_cache=kv_cache, is_pad_mask=is_pad_memory_mask)[0]
		x = x + self.mlp(self.mlp_ln(x))
		
		return x

@tables.register("decoder_classes", "SenseVoiceDecoder")
class SenseVoiceDecoder(nn.Module):
	def __init__(
		self, n_vocab: int, n_ctx: int, n_state: int, n_head: int, n_layer: int, **kwargs
	):
		super().__init__()
		
		self.token_embedding = nn.Embedding(n_vocab, n_state)
		self.positional_embedding = nn.Parameter(torch.empty(n_ctx, n_state))


		self.blocks = nn.ModuleList(
			[
				ResidualAttentionBlockRWKV(n_state, n_head, cross_attention=True, layer_id=i, **kwargs)
				for i in range(n_layer)
			]
		)
		self.ln = LayerNorm(n_state)
		
		mask = torch.empty(n_ctx, n_ctx).fill_(-np.inf).triu_(1)
		self.register_buffer("mask", mask, persistent=False)
		
		self.use_padmask = kwargs.get("use_padmask", True)


	
	def forward(
		self,
		x: torch.Tensor,
		xa: torch.Tensor,
		kv_cache: Optional[dict] = None,
		**kwargs,
	):
		"""Forward decoder.
	
		Args:
			hs_pad: encoded memory, float32  (batch, maxlen_in, feat)
			hlens: (batch)
			ys_in_pad:
				input token ids, int64 (batch, maxlen_out)
				if input_layer == "embed"
				input tensor (batch, maxlen_out, #mels) in the other cases
			ys_in_lens: (batch)
		Returns:
			(tuple): tuple containing:
	
			x: decoded token score before softmax (batch, maxlen_out, token)
				if use_output_layer is True,
			olens: (batch, )
		"""
		# import pdb;pdb.set_trace()
		use_padmask = self.use_padmask
		hlens = kwargs.get("hlens", None)
		
		ys_in_lens = kwargs.get("ys_in_lens", None)
		
		offset = next(iter(kv_cache.values())).shape[1] if kv_cache else 0
		tgt, memory = x, xa
		tgt[tgt == -1] = 0
		tgt = (
			self.token_embedding(tgt)
			+ self.positional_embedding[offset: offset + tgt.size(1)]
		)
		# tgt = self.dropout(tgt)
		
		x = tgt.to(memory.dtype)
		
		if use_padmask and hlens is not None:
			memory_mask = (~make_pad_mask(hlens)[:, None, :]).to(memory.device)
		else:
			memory_mask = None
		
		for layer, block in enumerate(self.blocks):
			x = block(x, memory, mask=self.mask, memory_mask=memory_mask, is_pad_mask=False, is_pad_memory_mask=True)
		
		x = self.ln(x)
		x = (
			x @ torch.transpose(self.token_embedding.weight.to(x.dtype), 0, 1)
		).float()
		
		return x



########################################################################################################
# The RWKV Language Model - https://github.com/BlinkDL/RWKV-LM
########################################################################################################

import os, math, gc, importlib
import torch
# torch._C._jit_set_profiling_executor(True)
# torch._C._jit_set_profiling_mode(True)
import torch.nn as nn
from torch.nn import functional as F



def __nop(ob):
	return ob


MyModule = nn.Module
MyFunction = __nop
if "RWKV_JIT_ON" in os.environ and os.environ["RWKV_JIT_ON"] == "1":
	MyModule = torch.jit.ScriptModule
	MyFunction = torch.jit.script_method

########################################################################################################
# CUDA Kernel
########################################################################################################

wkv6_cuda = None

def load_rwkv_kernel(HEAD_SIZE: int=64, RWKV_CTXLEN: int=512,):
	from torch.utils.cpp_extension import load
	global wkv6_cuda
	
	
	if wkv6_cuda is not None:
		return
	
	absolute_file_path = os.path.abspath(__file__)
	cur_dir = os.path.dirname(absolute_file_path)
	wkv6_cuda = load(name="wkv6", sources=[f"{cur_dir}/cuda/wkv6_op.cpp", f"{cur_dir}/cuda/wkv6_cuda.cu"],
	                 verbose=True, extra_cuda_cflags=["-res-usage", "--use_fast_math", "-O3", "-Xptxas -O3",
	                                                  "--extra-device-vectorization", f"-D_N_={HEAD_SIZE}",
	                                                  f"-D_T_={RWKV_CTXLEN}"])

# dtype = torch.float
dtype = torch.bfloat16
class WKV_6(torch.autograd.Function):
	@staticmethod
	def forward(ctx, B, T, C, H, r, k, v, w, u):
		with torch.no_grad():
			# assert r.dtype == torch.bfloat16
			# assert k.dtype == torch.bfloat16
			# assert v.dtype == torch.bfloat16
			# assert w.dtype == torch.bfloat16
			# assert u.dtype == torch.bfloat16
			# assert HEAD_SIZE == C // H
			ctx.B = B
			ctx.T = T
			ctx.C = C
			ctx.H = H
			assert r.is_contiguous()
			assert k.is_contiguous()
			assert v.is_contiguous()
			assert w.is_contiguous()
			assert u.is_contiguous()
			ew = (-torch.exp(w.float())).contiguous()
			ctx.save_for_backward(r, k, v, ew, u)
			y = torch.empty((B, T, C), device=r.device, dtype=dtype,
			                memory_format=torch.contiguous_format)  # .uniform_(-100, 100)
			wkv6_cuda.forward(B, T, C, H, r, k, v, ew, u, y)
			return y
	
	@staticmethod
	def backward(ctx, gy):
		with torch.no_grad():
			# assert gy.dtype == torch.bfloat16
			B = ctx.B
			T = ctx.T
			C = ctx.C
			H = ctx.H
			assert gy.is_contiguous()
			r, k, v, ew, u = ctx.saved_tensors
			gr = torch.empty((B, T, C), device=gy.device, requires_grad=False, dtype=dtype,
			                 memory_format=torch.contiguous_format)  # .uniform_(-100, 100)
			gk = torch.empty((B, T, C), device=gy.device, requires_grad=False, dtype=dtype,
			                 memory_format=torch.contiguous_format)  # .uniform_(-100, 100)
			gv = torch.empty((B, T, C), device=gy.device, requires_grad=False, dtype=dtype,
			                 memory_format=torch.contiguous_format)  # .uniform_(-100, 100)
			gw = torch.empty((B, T, C), device=gy.device, requires_grad=False, dtype=dtype,
			                 memory_format=torch.contiguous_format)  # .uniform_(-100, 100)
			gu = torch.empty((B, C), device=gy.device, requires_grad=False, dtype=dtype,
			                 memory_format=torch.contiguous_format)  # .uniform_(-100, 100)
			wkv6_cuda.backward(B, T, C, H, r, k, v, ew, u, gy, gr, gk, gv, gw, gu)
			gu = torch.sum(gu, 0).view(H, C // H)
			return (None, None, None, None, gr, gk, gv, gw, gu)


def RUN_CUDA_RWKV6(B, T, C, H, r, k, v, w, u):
	return WKV_6.apply(B, T, C, H, r, k, v, w, u)


class RWKV_Tmix_x060(MyModule):
	def __init__(self, args, layer_id):
		super().__init__()
		self.args = args
		
		load_rwkv_kernel(args.head_size_a, args.ctx_len)
		
		self.layer_id = layer_id
		
		self.head_size = args.head_size_a
		self.n_head = args.dim_att // self.head_size
		assert args.dim_att % self.n_head == 0
		
		with torch.no_grad():
			ratio_0_to_1 = layer_id / (args.n_layer - 1)  # 0 to 1
			ratio_1_to_almost0 = 1.0 - (layer_id / args.n_layer)  # 1 to ~0
			ddd = torch.ones(1, 1, args.n_embd)
			for i in range(args.n_embd):
				ddd[0, 0, i] = i / args.n_embd
			
			# fancy time_mix
			self.time_maa_x = nn.Parameter(1.0 - torch.pow(ddd, ratio_1_to_almost0))
			self.time_maa_w = nn.Parameter(1.0 - torch.pow(ddd, ratio_1_to_almost0))
			self.time_maa_k = nn.Parameter(1.0 - torch.pow(ddd, ratio_1_to_almost0))
			self.time_maa_v = nn.Parameter(1.0 - (torch.pow(ddd, ratio_1_to_almost0) + 0.3 * ratio_0_to_1))
			self.time_maa_r = nn.Parameter(1.0 - torch.pow(ddd, 0.5 * ratio_1_to_almost0))
			self.time_maa_g = nn.Parameter(1.0 - torch.pow(ddd, 0.5 * ratio_1_to_almost0))
			
			D_MIX_LORA = 32  # generate TIME_MIX for w,k,v,r,g
			self.time_maa_w1 = nn.Parameter(torch.zeros(args.n_embd, D_MIX_LORA * 5))
			self.time_maa_w2 = nn.Parameter(torch.zeros(5, D_MIX_LORA, args.n_embd).uniform_(-0.01, 0.01))
			
			# fancy time_decay
			decay_speed = torch.ones(args.dim_att)
			for n in range(args.dim_att):
				decay_speed[n] = -6 + 5 * (n / (args.dim_att - 1)) ** (0.7 + 1.3 * ratio_0_to_1)
			self.time_decay = nn.Parameter(decay_speed.reshape(1, 1, args.dim_att))
			
			D_DECAY_LORA = 64
			self.time_decay_w1 = nn.Parameter(torch.zeros(args.n_embd, D_DECAY_LORA))
			self.time_decay_w2 = nn.Parameter(torch.zeros(D_DECAY_LORA, args.dim_att).uniform_(-0.01, 0.01))
			
			tmp = torch.zeros(args.dim_att)
			for n in range(args.dim_att):
				zigzag = ((n + 1) % 3 - 1) * 0.1
				tmp[n] = ratio_0_to_1 * (1 - (n / (args.dim_att - 1))) + zigzag
			
			self.time_faaaa = nn.Parameter(tmp.reshape(self.n_head, self.head_size))
		
		self.time_shift = nn.ZeroPad2d((0, 0, 1, -1))
		self.receptance = nn.Linear(args.n_embd, args.dim_att, bias=False)
		self.key = nn.Linear(args.n_embd, args.dim_att, bias=False)
		
		self.value = nn.Linear(args.n_embd, args.dim_att, bias=False)
		self.output = nn.Linear(args.dim_att, args.n_embd, bias=False)
		self.gate = nn.Linear(args.n_embd, args.dim_att, bias=False)
		self.ln_x = nn.GroupNorm(self.n_head, args.dim_att, eps=(1e-5) * (args.head_size_divisor ** 2))
	
	@MyFunction
	def jit_func(self, x):
		B, T, C = x.size()
		
		xx = self.time_shift(x) - x
		
		xxx = x + xx * self.time_maa_x
		xxx = torch.tanh(xxx @ self.time_maa_w1).view(B * T, 5, -1).transpose(0, 1)
		xxx = torch.bmm(xxx, self.time_maa_w2).view(5, B, T, -1)
		mw, mk, mv, mr, mg = xxx.unbind(dim=0)
		
		xw = x + xx * (self.time_maa_w + mw)
		xk = x + xx * (self.time_maa_k + mk)
		xv = x + xx * (self.time_maa_v + mv)
		xr = x + xx * (self.time_maa_r + mr)
		xg = x + xx * (self.time_maa_g + mg)
		
		r = self.receptance(xr)
		k = self.key(xk)
		v = self.value(xv)
		g = F.silu(self.gate(xg))
		
		ww = torch.tanh(xw @ self.time_decay_w1) @ self.time_decay_w2
		w = self.time_decay + ww
		
		return r, k, v, g, w
	
	@MyFunction
	def jit_func_2(self, x, g):
		B, T, C = x.size()
		x = x.view(B * T, C)
		
		x = self.ln_x(x).view(B, T, C)
		x = self.output(x * g)
		return x
	
	def forward(self, x):
		B, T, C = x.size()
		H = self.n_head
		
		r, k, v, g, w = self.jit_func(x)
		x = RUN_CUDA_RWKV6(B, T, C, H, r, k, v, w, u=self.time_faaaa)
		
		return self.jit_func_2(x, g)

class RWKV_CMix_x060(MyModule):
	def __init__(self, args, layer_id):
		super().__init__()
		self.args = args
		self.layer_id = layer_id
		self.time_shift = nn.ZeroPad2d((0, 0, 1, -1))
		
		with torch.no_grad():  # fancy init of time_mix
			ratio_1_to_almost0 = 1.0 - (layer_id / args.n_layer)  # 1 to ~0
			ddd = torch.ones(1, 1, args.n_embd)
			for i in range(args.n_embd):
				ddd[0, 0, i] = i / args.n_embd
			self.time_maa_k = nn.Parameter(1.0 - torch.pow(ddd, ratio_1_to_almost0))
			self.time_maa_r = nn.Parameter(1.0 - torch.pow(ddd, ratio_1_to_almost0))
		
		self.key = nn.Linear(args.n_embd, args.dim_ffn, bias=False)
		self.receptance = nn.Linear(args.n_embd, args.n_embd, bias=False)
		self.value = nn.Linear(args.dim_ffn, args.n_embd, bias=False)
	
	@MyFunction
	def forward(self, x):
		xx = self.time_shift(x) - x
		xk = x + xx * self.time_maa_k
		xr = x + xx * self.time_maa_r
		
		k = self.key(xk)
		k = torch.relu(k) ** 2
		kv = self.value(k)
		return torch.sigmoid(self.receptance(xr)) * kv


class Block(nn.Module):
	def __init__(self, args, layer_id):
		super().__init__()
		self.args = args
		self.layer_id = layer_id
		
		self.ln1 = nn.LayerNorm(args.n_embd)
		self.ln2 = nn.LayerNorm(args.n_embd)
		
		if self.layer_id == 0:
			self.ln0 = nn.LayerNorm(args.n_embd)


		self.att = RWKV_Tmix_x060(args, layer_id)
		
		self.ffn = RWKV_CMix_x060(args, layer_id)


		if args.dropout > 0:
			self.drop0 = nn.Dropout(p=args.dropout)
			self.drop1 = nn.Dropout(p=args.dropout)
	
	def forward(self, x, x_emb=None):
		args = self.args
		B, T, C = x.size()
		if self.layer_id == 0:
			x = self.ln0(x)

		
		if self.args.dropout == 0:
			if self.layer_id == 0 and args.pre_ffn > 0:
				x = x + self.ffnPre(self.ln1(x))
			else:
				x = x + self.att(self.ln1(x))
			x = x + self.ffn(self.ln2(x))
		else:
			if self.layer_id == 0 and args.pre_ffn > 0:
				x = self.drop0(x + self.ffnPre(self.ln1(x)))
			else:
				x = self.drop0(x + self.att(self.ln1(x)))
			x = self.drop1(x + self.ffn(self.ln2(x)))
		
		return x


class RWKVLayer(nn.Module):
	def __init__(self, args, layer_id):
		super().__init__()
		self.args = args
		self.layer_id = layer_id
		if args.dim_ffn is None:
			args.dim_ffn = int((args.n_embd * 3.5) // 32 * 32)
		self.ln0 = None
		if self.layer_id == 0 and args.get("ln0", True):
			self.ln0 = nn.LayerNorm(args.n_embd)
		
		self.ln1 = None
		if args.get("ln1", True):
			self.ln1 = nn.LayerNorm(args.n_embd)
		self.ln2 = nn.LayerNorm(args.n_embd)
		

		self.att = RWKV_Tmix_x060(args, layer_id)
		
		self.ffn = RWKV_CMix_x060(args, layer_id)
		
		if args.dropout > 0:
			self.drop0 = nn.Dropout(p=args.dropout)
			self.drop1 = nn.Dropout(p=args.dropout)
		
		# init
		if args.get("init_rwkv", True):
			print("init_rwkv")
			nn.init.orthogonal_(self.att.receptance.weight, gain=1)
			nn.init.orthogonal_(self.att.key.weight, gain=0.1)
			nn.init.orthogonal_(self.att.value.weight, gain=1)
			nn.init.orthogonal_(self.att.gate.weight, gain=0.1)
			nn.init.zeros_(self.att.output.weight)
			
			nn.init.orthogonal_(self.ffn.key.weight, gain=1)
			nn.init.zeros_(self.ffn.value.weight)
			nn.init.zeros_(self.ffn.receptance.weight)
			scale = ((1 + layer_id) / args.get("n_layer")) ** 0.7
			nn.init.constant_(self.ln2.weight, scale)
			if self.ln0 is not None:
				nn.init.constant_(self.ln0.weight, scale)
			if self.ln1 is not None:
				nn.init.constant_(self.ln1.weight, scale)
		
	def forward(self, x, x_emb=None, mask=None, **kwargs):
		
		args = self.args
		if args.get("datatype", "bf16") == "bf16":
			x = x.bfloat16()
		B, T, C = x.size()
		if self.layer_id == 0 and self.ln0 is not None:
			x = self.ln0(x)
		
		if self.args.dropout == 0:
			if self.ln1 is None:
				x = x + self.att(x)
			else:
				x = x + self.att(self.ln1(x))
			x = x + self.ffn(self.ln2(x))
		else:
			if self.ln1 is None:
				x = self.drop0(x + self.att(x))
			else:
				x = self.drop0(x + self.att(self.ln1(x)))
			x = self.drop1(x + self.ffn(self.ln2(x)))
		
		if args.get("datatype", "bf16") == "bf16":
			x = x.to(torch.float32)
		return x


class RWKV(nn.Module):
	def __init__(self, args):
		super().__init__()
		self.args = args
		if not hasattr(args, 'dim_att'):
			args.dim_att = args.n_embd
		if not hasattr(args, 'dim_ffn'):
			if '-f4' in os.environ["RWKV_MY_TESTING"]:
				args.dim_ffn = int((args.n_embd * 4) // 32 * 32)
			else:
				args.dim_ffn = int((args.n_embd * 3.5) // 32 * 32)  # default = 3.5x emb size
		if not hasattr(args, 'tiny_att_layer'):
			args.tiny_att_layer = -1
		if not hasattr(args, 'tiny_att_dim'):
			args.tiny_att_dim = -1
		assert args.n_embd % 32 == 0
		assert args.dim_att % 32 == 0
		assert args.dim_ffn % 32 == 0
		
		self.emb = nn.Embedding(args.vocab_size, args.n_embd)
		
		self.blocks = nn.ModuleList([Block(args, i) for i in range(args.n_layer)])
		
		self.ln_out = nn.LayerNorm(args.n_embd)
		self.head = nn.Linear(args.n_embd, args.vocab_size, bias=False)


		if args.dropout > 0:
			self.drop0 = nn.Dropout(p=args.dropout)


	def forward(self, idx):
		args = self.args
		B, T = idx.size()
		assert T <= args.ctx_len, "Cannot forward, model ctx_len is exhausted."
		
		x = self.emb(idx)
		x_emb = x
		
		if args.dropout > 0:
			x = self.drop0(x)
		if args.tiny_att_dim > 0:
			for block in self.blocks:
				if args.grad_cp == 1:
					x = deepspeed.checkpointing.checkpoint(block, x, x_emb)
				else:
					x = block(x, x_emb)
		else:
			for block in self.blocks:
				if args.grad_cp == 1:
					x = deepspeed.checkpointing.checkpoint(block, x)
				else:
					x = block(x)
		
		x = self.ln_out(x)
		
		if args.head_qk > 0:
			q = self.head_q(x)[:, :T, :]
			k = self.head_k(x)[:, :T, :]
			c = (q @ k.transpose(-2, -1)) * (1.0 / args.head_qk)
			c = c.masked_fill(self.copy_mask[:T, :T] == 0, 0)
			
			if "32" in os.environ["RWKV_FLOAT_MODE"]:
				c = c @ F.one_hot(idx, num_classes=args.vocab_size)
			elif os.environ["RWKV_FLOAT_MODE"] == "fp16":
				c = c @ F.one_hot(idx, num_classes=args.vocab_size).half()
			elif os.environ["RWKV_FLOAT_MODE"] == "bf16":
				c = c @ F.one_hot(idx, num_classes=args.vocab_size).bfloat16()
			
			x = self.head(x) + c
		else:
			x = self.head(x)
		
		return x
